/* define the list management functions */

#include <stdlib.h>

#include "reporter.h"
#include "list.h"

struct ListNode {
  struct ListNode *Next;
  ElementType Element;
};

/* Return a pointer to a ListNode (the header) */

List CreateList(void) {
  List tmp;
  tmp = malloc(sizeof(struct ListNode));
  FatalErrorMemory(tmp);
  tmp->Next = NULL;
  return tmp;
}

/* Release all memory associated with the list including the header node. */

void DestroyList(List L) {
  if (L == NULL) return;
  ClearList(L);
  free(L);
}

/*
Return a pointer to an array of pointers to ListNodes (headers).
This procedure allocates all the header nodes at once, rather than
individuall in repeated calls to malloc.
*/

List *CreateListArray(int n) {
  List *list;
  struct ListNode *header;
  int i;

  list = malloc(n*sizeof(List));
  FatalErrorMemory(list);
  header = malloc(n*sizeof(struct ListNode));
  FatalErrorMemory(header);
  for (i=0; i<n; i++) {
    header[i].Next = NULL;
    list[i] = &header[i];
  }
  return list;
}

/*
Release all memory associated with an array of lists, including the
header nodes.
*/
void DestroyListArray(List *LA, int n) {
  int i;
  if (LA == NULL) return;
  if (n<=0) return;
  for (i=0; i<n; i++)
    ClearList(LA[i]);
  free(LA[0]);
  free(LA);
  return;
}

/*
Release all the list nodes in the list except header node.
Note that this does not release the memory occupied by the
Elements stored in the list.  The caller needs to do this
independently somehow.  Tacky, but simpler than making 
provisions for object destructors ...
*/

void ClearList(List L) {
  Position tmp;
  Position cur;

  if (L == NULL) return;
  cur = L->Next;
  while (cur != NULL) {
    tmp = cur->Next;
    free(cur);
    cur = tmp;
  }
  L->Next = NULL;
}
 
void PrintList(List L, ElementPrintLabel P) {
  Position current;
  current = L;
  while ((current = current->Next) != NULL) {
    (*P)(current->Element);
    printf(" ");
  }
}

int CountListEntries(List L) {
  Position current;
  int count;

  if (L==NULL) return 0;

  count = 0;
  current = L->Next;

  while (current != NULL) {
    count++;
    current = current->Next;
  }
  return count;
}

Position FindListEntry(List L,ElementType X,Comparator C) {
  Position cur;
  
  if (L == NULL) return NULL;

  cur = L->Next;
  while (cur != NULL && (*C)(cur->Element,X) != 0)
    cur = cur->Next;
  return cur;
}

Position FindPreviousListEntry(List L, ElementType X, Comparator C) {
  Position P;

  if (L==NULL) return NULL;

  P = L;
  while (P->Next != NULL && (*C)(P->Next->Element,X) != 0)
    P = P->Next;
  return P;
}

Position InsertListEntry(List L, Position P, ElementType X) {
  Position tmp;
  tmp = malloc(sizeof(struct ListNode));
  FatalErrorMemory(tmp);
  tmp->Element = X;
  tmp->Next = P->Next;
  P->Next = tmp;
  return tmp;
}

void DeleteListEntry(List L, Position P) {
  Position tmp;

  tmp = L;
  while (tmp != NULL) {
    if (tmp->Next == P) break;
    tmp = tmp->Next;
  }
  if (tmp != NULL) {
    tmp->Next = P->Next;
    free(P);
  }
}

int ListIsEmpty(List L) {
  return (L->Next == NULL);
}

Position GetFirstListEntry(List L) {
  return L->Next;
}

ElementType GetListEntryElement(Position P) {
  return P->Element;
}

Position AdvanceListEntry(Position P) {
  return P->Next;
}

int ListEntryIsLast(List L,Position P) {
  return (P->Next == NULL);
}
