#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "symbol.h"
#include "reporter.h"

/* 

This package defines some procedures for dealing with simple Symbol
objects.  The symbols are defined using text files.

The format of each row of the input symbol table file is simply

  <numeric value> <symbol name>

An example of this format can be produced in unix with the following command.

 nm --defined-only -n <exe> | grep " T " | cut -d ' ' -f 1,3 > symbols.txt

Other possibilities include

 ps -eo pid,comm --no-headers > pidsymbols.txt
 ls -i1 /usr/bin > binsymbols.txt
 ls -iR /usr | grep  -v '^$\|\:' > usrsymbols.txt

Read in all the symbol definitions and create a little object for each
one.  Return an array of pointers to the objects, terminated by a NULL
pointer.

If there is an allocation error, this procedure terminates the calling
process immediately.

*/

/*-----------------------------------------------------------------*/

struct Symbol **readSymbols(FILE *src) {

#define INITIAL_CAPACITY 100	/* initial size of pointer buffer */

  long capacity;		/* capacity of array */
  long count;			/* number of symbols provided */
  long maxcap;			/* maximum number of symbols allowed */

  struct Symbol **A;		/* pointer to array of object pointers */
  struct Symbol **B;		/* pointer to array of object pointers */

  char fmt[100];		/* format string for decoding input */
#define MAX_NAME_LENGTH 100
  char symName[MAX_NAME_LENGTH+1]; /* symbol name */
  unsigned long symValue;	/* symbol value */

  struct Symbol *symP;		/* pointer to a Symbol object */

  /* 
  "size_t" must be at least as big as "long" for easier limit
   calculations.
  */
  FatalErrorBound(sizeof(long),sizeof(size_t));
  maxcap = LONG_MAX/sizeof(struct Symbol *);

  capacity = INITIAL_CAPACITY;
  FatalErrorBound(capacity,maxcap);
  A = malloc(capacity*sizeof(struct Symbol *));
  FatalErrorMemory(A);

  count = 0;
  sprintf(fmt,"%%8lx%%%is",MAX_NAME_LENGTH);
  while (fscanf(src,fmt,&symValue,symName) == 2) {

    symP = malloc(sizeof(struct Symbol));
    FatalErrorMemory(symP);

    symP->name = malloc((strlen(symName)+1)*sizeof(char));
    FatalErrorMemory(symP->name);
    strcpy(symP->name,symName);

    symP->value = symValue;

    if (++count >= capacity) {	/* array is full, reallocate */
      FatalErrorBound(capacity,(maxcap/2));
      capacity *= 2;
      B = realloc(A,capacity*sizeof(struct Symbol *));
      FatalErrorMemory(B);
      A = B;
    }
    A[count-1] = symP;
  }
  A[count] = NULL;
  return A;
}

/* This function assumes that all the Symbols in the table still exist. */

void freeSymbols(struct Symbol **table) {
  long count;
  struct Symbol *symP;
  if (table == NULL) return;
  count = 0;
  while (table[count]!=NULL) {
    symP = (struct Symbol *)table[count];
    if (symP->name != NULL) {
      free(symP->name);
    }
    free(symP);
    count++;
  }
}

int symbolCompareByName(void *a, void *b) {
  return strcmp(((struct Symbol *)a)->name,((struct Symbol *)b)->name);
}

int symbolCompareByValue(void *a, void *b) {
  if (((struct Symbol *)a)->value < ((struct Symbol *)b)->value) 
    return -1;
  else if (((struct Symbol *)a)->value > ((struct Symbol *)b)->value)
    return +1;
  else
    return 0;
}
